/*
 * generic.c: program to control leds and pushbutton switch
 * on a WRAP board
 *
 * portions based on
 * (c) 2005 by Torsten Georg <fli4l<at>foddy.de>
 * 
 * $Id: embd.c 15248 2008-07-21 10:15:14Z jw5 $
 * 
 * embd reads from a fifo file commands to execute.
 * valid commands are: led1|led2|led3 on|off|blink
 *
 * (c) Copyright 2006 Stefan Krister <stefan@fli4l.de>
 * and the fli4l developer-team, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 * 
 * I don't accept liability nor provide warranty for this Software.
 * This material is provided "AS-IS" and at no charge.
 * 
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/io.h>

#include <linux/kd.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <syslog.h>

#include "embd.h"

// buffer for fifo
#define BUFFERSIZE 1024

int
usage(char * argv)
{
    printf("--------------------------------------------------------------\n");
    printf(" embd usage\n");
    printf("\n");
    printf(" Usage: %s [-d] [-f] [-i] [-b] [ -n <num leds> ] [ -p <fifo> ] [ -t <type> command [ arg1, ..., argn ]\n", argv);
    printf("--------------------------------------------------------------\n");
    exit (1);
}

int embd_invert = 0;
int embd_no_button = 0;
int embd_max_leds = MAX_LEDS;
struct led_vec led_vectors[MAX_LEDS] = { 
    {0, 0x04, GPIOBASE, -1}, 
    {0, 0x08, GPIOBASE, -1}, 
    {0, 0x04, GPIOBASE+2, -1}};

struct button {
    int status;
    int fd;
} button = { 1, -1};

void (* switch_led)(int num, int on);
int (* read_button)(void);
int soft_button;

int daemonized = 0;
int fg = 0;
int debug = 0;
int fd = -1;

void dump_led_status(int num) 
{
    dolog(LOG_INFO, "LED%d %s %s", num+1, 
	  led_vectors[num].status & 1 ? "on " : "off",
	  led_vectors[num].status & 2 ? "(blinking)" : "");
}

void dump_switch_status(void)
{
    dolog(LOG_INFO,"status: button %s", 
	  button.status ? "released" : "pressed" );
}

void dump_status(void)
{
    int i;
    for (i=0; i<embd_max_leds; i++)
	dump_led_status(i);
    dump_switch_status();
}

/** dev controlled led
 */
#define LED_DEV "/dev/embdled1"
char led_dev_tmpl[32] = LED_DEV;
char * const led_dev_index = led_dev_tmpl + sizeof(LED_DEV) - 2;
const char button_dev[] = "/dev/embdbutton";

void switch_led_via_dev(int num, int on)
{
    int fd;
    unsigned char val = on ? '0' : '1';
    
    *led_dev_index = '1' + num;
    if ( (fd = open(led_dev_tmpl, O_RDWR)) < 0) {
        dolog(LOG_ERR, "error opening device %s: %s",
	      led_dev_tmpl, strerror(errno));
        return;
    }
    if (write(fd, &val, sizeof(val)) < 0) {
	dolog(LOG_ERR, "error writing led device: %s",
	      strerror(errno));
    }
    close(fd);
}

int read_button_via_dev(void)
{
    int fd;
    unsigned char state;
    
    if ( (fd = open(button_dev, O_RDONLY)) < 0) {
	dolog(LOG_ERR, "error opening device %s: %s, ignoring button",
	      button_dev, strerror(errno));
	read_button = 0;
	return button.status;
    }

    if (read(fd, &state, sizeof(state)) < 0) {
        close(fd);
	dolog(LOG_ERR, "error reading button device: %s, ignoring button",
	      strerror(errno));
	read_button = 0;
	return button.status;
    }
    
    close(fd);
    return state == '1';
}

void dev_led_init(void)
{
    switch_led = switch_led_via_dev;

    if (embd_no_button) {
	read_button = 0;
    } 
    else if (access(button_dev, R_OK) < 0) {
        dolog(LOG_ERR, "Unable to access button device %s (%s), ignoring button", 
	      button_dev, strerror(errno));
	read_button = 0;
    }
    else {
       read_button = read_button_via_dev;
    }
}

/** keyboard leds
 */
void switch_keyboard_led(int num, int on)
{
    int fd = led_vectors[0].fd;
    unsigned char val;
    int i;

    if (fd < 0)
	return;

    for (val=i=0; i<3; i++)
	if (led_vectors[i].status & 1)
	    val |= (1<<i);

    if (ioctl(fd, KDSETLED, val) < 0) {
	dolog(LOG_ERR, "error setting led state: %s\n", strerror(errno));
    }
}

void keyboard_led_init(void)
{
    char * name = "/dev/console";
    if ( (led_vectors[0].fd = open(name, O_RDONLY)) < 0) 
	dolog(LOG_ERR, "error opening device %s: %s",
	      name, strerror(errno));
    switch_led = switch_keyboard_led;
    read_button = 0;
}

/** simulated led
 */
void switch_led_simulation(int num, int on)
{
    dump_led_status(num);
}

int soft_button = 1;
int read_button_simulation(void)
{
    return soft_button;
}

void simulation_init(void)
{
    switch_led = switch_led_simulation;
    read_button = read_button_simulation;
}

char * fifo_name = "/var/run/embedded.fifo";


#define verify_arg(num) do { if (num < 0 || num >= embd_max_leds) return; } while(0)

void set_led(int num, int status)
{
    int on;
    verify_arg(num);
    led_vectors[num].status = status;
    on = status & 1;
    switch_led(num, embd_invert ? !on : on);
}

void toggle_led(int num)
{
    verify_arg(num);
    if (led_vectors[num].status < 2)
	return;

    if (led_vectors[num].status & 1)
	set_led(num, led_vectors[num].status & ~1U);
    else
	set_led(num, led_vectors[num].status | 1);
}

void enable_blink(int num)
{
    verify_arg(num);
    led_vectors[num].status |= 2;
}

int pid; 
void execute_command (char ** command)
{
    int res; 

    if (pid)
        return;

    pid = fork ();
    switch (pid) {
    case -1:
        dolog(LOG_ERR, "unable to fork: %s", strerror(errno));
	pid = 0;
    default:
	return;
    case 0:
        close(fd);
	res = execvp(command[0], command);
        dolog(LOG_ERR, "unable to exec %s: %s", command[0], strerror(errno));
	exit(res);
    }
}

inline void check_for_zombies (void)
{
    int status, res;

    if (!pid)
	return;

    res = waitpid(-1, &status, WNOHANG);
    if (res < 0) {
	dolog(LOG_ERR, "waitpid returned %s\n", strerror(errno));
    } else if (res > 0) {
	if (WIFEXITED(status)) {
	    dolog(LOG_INFO, "%d terminated with %d\n", 
		  res, WEXITSTATUS(status));
	} else {
	    dolog(LOG_INFO, "%d terminated abnormal\n", res);
	}
	pid = 0;
    }
}

void handle_button (char ** command)
{
    // read status of pushbutton switch
    // 1 = released
    // 0 = pressed

    int old_status;

    if (embd_no_button || !read_button)
	return;

    old_status = button.status; 
    button.status = read_button();
    if ( old_status == 1 && button.status == 0)
    {
	dolog(LOG_INFO, "button pressed, executing command %s", command[0]);
	execute_command(command);
    }
}

int fd;
void open_fifo()
{
    fd = open(fifo_name, O_RDONLY | O_NONBLOCK);
    if(fd == -1)
    {
	dolog(LOG_ERR, "could not open FIFO file");
	exit (1);
    }
}

int handle_fifo(int fd)
{
    char   buf[BUFFERSIZE]; 
    int i=0, res;

    for (i=0; i<BUFFERSIZE-1; i++) {
	res = read(fd, &buf[i], sizeof(char));
	if (res < 0) {
	    dolog(LOG_ERR, "error reading fifo: %m\n");
	    break;
	}
	if (res == 0 || buf[i] == '\n')
	    break;
    }
    buf[i]='\0';

    if (i > 0)
    {
	char token[11], mode[11];
	char led_token[] = "led";

	dolog(LOG_DEBUG, "gelesen: \"%s\"", buf);
	
	if (sscanf(buf, " %10s %10s ", token, mode) < 1) {
	    dolog(LOG_ERR, "invalid request '%s'", buf);
	    return 1;
	}

	if (!strcasecmp(token, "status")) {
	    dump_status();
	    return 1;
	}
	if (!strcasecmp(token, "button")) {
	    soft_button = (*mode == 'u' || *mode == 'U') ? 1 : 0;
	    return 1;
	}
	if (!strncasecmp(token, led_token, sizeof(led_token)-1)) {
	    int num = token[3] - '1';
	    if (!strcasecmp(mode, "on"))
		set_led(num, 1);
	    else if (!strcasecmp(mode, "off"))
		set_led(num, 0);
	    else if (!strcasecmp(mode, "blink"))
		enable_blink(num);
	}
    } 
    else
    {
	close(fd);
	open_fifo();
	return 0;
    }
    return 1;
}
/*****************************************************************************/
/*                                                                           */
/*                         M A I N   P R O G R A M                           */
/*                                                                           */
/*****************************************************************************/
int
main (int argc, char **argv)
{
    char ** command;
    char * type = "simulation";

    int res, idx, fg = 0;
    char c;

    while ((c = getopt (argc, argv, "fdibn:p:t:")) != -1)
	switch (c) {
	case 'f':
	    fg = 1;
	    break;
	case 'd':
	    debug = 1;
	    break;
	case 'p':
	    fifo_name = optarg;
	    break;
	case 't':
	    type = optarg;
	    break;
	case 'n':
	    embd_max_leds = atoi(optarg);
	    if (embd_max_leds > MAX_LEDS)
	        embd_max_leds = MAX_LEDS;
	    break;
	case 'i':
	    embd_invert = 1;
	    break;
	case 'b':
	    embd_no_button = 1;
	    break;
	default:
	    usage (argv[0]);
	}
    if (optind + 1 > argc) 
	usage (argv[0]);

    command = argv + optind;

    // Delete (old) FIFO file
    unlink(fifo_name);

    // Create FIFO file
    res = mkfifo(fifo_name, 0777);
    if(res != 0)
    {
	dolog(LOG_ERR, "could not create FIFO file");
	exit (1);
    }
    open_fifo();
    
    switch (*type) {
    case 'h': 
	direct_led_init();
	break;
    case 'd':
	dev_led_init();
	break;
    case 'p':
	keyboard_led_init();
	break;
    case 'j':
	jnt_led_init();
	break;
    default:
      dolog(LOG_ERR, "unknown type '%s', falling back to simulation", type);
      /* fall through for the time beeing */
    case 's':
	simulation_init();
	break;
    }

    if (!fg) {
	if (daemon(1, 1) == -1)
	{
	    dolog(LOG_ERR, "cannot detach");
	    exit(1);
	}
	daemonized = 1;
    } 
    openlog("embd", LOG_PID, LOG_USER);
    
    for (idx=0; idx<embd_max_leds; idx++)
	switch_led(idx, 0);
        
    while (1) {
	struct timeval tv;
	fd_set fds;

	tv = (struct timeval){1, 0};
	FD_ZERO(&fds);
	FD_SET(fd,&fds);
	
	if ( (res = select(fd+1, &fds, 0, 0, &tv)) < 0) {
	    dolog(LOG_ERR, "error in select: %s", strerror(errno));
	    close(fd);
	    open_fifo();
	    continue;
	}

	if (res == 1)
	    if ( !handle_fifo(fd))
		continue;

	check_for_zombies();
	handle_button(command);

	// handle blinking LEDs
	for (idx=0; idx < embd_max_leds; idx++)
	    toggle_led(idx);
    }
    return 0;
}

